import json
import csv
import random
from collections import OrderedDict

from redleader.cluster import Cluster, AWSContext
from redleader.resources.elasticbeanstalk import *
from redleader.managers import ElasticBeanstalkManager
from redleader.util import sanitize
import redleader.resources as r

import pressurize.AWSManager

class ResourceManager(object):
    """
    ResourceManager coordinates AWS resources for
    a pressurize deployment
    """
    def __init__(self, controller):
        self._controller = controller

        # Our AWSManager
        self._aws_manager = self._controller._aws_manager

        # RedLeader AWSContext
        self._aws_context = AWSContext(
            aws_profile = self._controller._aws_profile,
            aws_region=self._controller.config['aws_region'])

    def _cluster_name(self):
        return self._controller.config["deployment_name"] + "Cluster"

    def prefix_name(self, name):
        return self._controller.config["deployment_name"] + name

    def default_dynamo_tables(self, context):
        """
        Create default DynamoDB tables
        """
        cache_config = {
            "key_schema": OrderedDict([
                ('key', 'HASH'),
            ]),
            "attribute_definitions": {
                'key': 'S'
            }
        }
        self.cache_table = r.DynamoDBTableResource(
            context, self.prefix_name("cache"),
            attribute_definitions=cache_config['attribute_definitions'],
            key_schema=cache_config['key_schema'],
            write_units=5, read_units=5
        )
        print("DynamoDB cache table name", self.prefix_name("cache"))

        auth_config = {
            "key_schema": OrderedDict([
                ('token_key', 'HASH'),
            ]),
            "attribute_definitions": {
                'token_key': 'S'
            }
        }
        self.auth_table = r.DynamoDBTableResource(
            context, self.prefix_name("auth"),
            attribute_definitions=auth_config['attribute_definitions'],
            key_schema=auth_config['key_schema'],
            write_units=5, read_units=5
        )
        print("DynamoDB auth table name", self.prefix_name("cache"))

        return [self.cache_table, self.auth_table]

    def bucket_resources(self, context, model):
        """
        Create RedLeader bucket resources for any referenced buckets,
        so that our deployed elastic beanstalk applications can be granted
        appropriate IAM permissions
        """
        bucket_names = {}
        dynamodb_table_names = {}
        for resource_name in model["required_resources"]:
            resource = model["required_resources"][resource_name]
            if("s3://" in resource):
                parts = resource.split("/")
                if len(parts) < 3:
                    raise RuntimeError("Invalid s3 url in configuration: %s" % resource)
                bucket_names[parts[2]] = True
            if("dynamodb://" in resource):
                parts = resource.split("//")
                if len(parts) < 2:
                    raise RuntimeError("Invalid dynamodb table in configuration: %s" % resource)
                dynamodb_table_names[parts[2]] = { "aws_region":
                                                   context._aws_region if len(parts) < 3 else parts[1] }

        bucket_resources = []
        for bucket_name in bucket_names:
            bucket_resources.append(r.S3BucketResource(context, bucket_name))

        table_resources = []
        for table_name in dynamodb_table_names:
            conf = dynamodb_table_names[table_name]
            print("Creating resource ", table_name, conf['aws_region'])
            bucket_resources.append(r.DynamoDBTableResource(context, table_name, aws_region=conf['aws_region']))

        return bucket_resources + table_resources

    def elastic_beanstalk_bucket(self):
        return "pressurizebucket" + sanitize(self._controller.config['deployment_name'])

    def cname_prefix(self, name):
        return sanitize(self._controller.config["deployment_name"]) + "-" + sanitize(name)

    def elastic_beanstalk_api_resources(self, version, source_file, config_options):
        """
        Create RedLeader resources for the API elastic beanstalk deployment
        """
        context = self._aws_context
        app = ElasticBeanstalkAppResource(context, self.prefix_name("api"))
        cname_prefix = self.cname_prefix("api")
        table_resources = self.default_dynamo_tables(context)
        permission_resources = list(map(lambda x: r.ReadWritePermission(x), table_resources))
        version = ElasticBeanstalkAppVersionResource(
            context,
            app,
            self.elastic_beanstalk_bucket(),
            source_file.split("/")[-1],
            version)
        config = ElasticBeanstalkConfigTemplateResource(
            context,
            app,
            config_options,
            solution_stacks["docker"],
            "Pressurize API beanstalk config",
            permission_resources=permission_resources
        )
        env = ElasticBeanstalkEnvResource(
            context,
            app,
            version,
            config,
            cname_prefix,
            "Pressurize API env"
        )
        return [app, version, config, env] + table_resources

    def elastic_beanstalk_model_resources(self, name, version, source_file, config_options):
        """
        Create RedLeader resources for a single elastic beanstalk model deployment
        """
        context = self._aws_context
        app = ElasticBeanstalkAppResource(context, self.prefix_name(sanitize(name)))
        cname_prefix = self.cname_prefix(name)
        model_config = self._controller.models[name]
        bucket_resources = self.bucket_resources(context, model_config)

        permission_resources = list(map(lambda x: r.ReadWritePermission(x), bucket_resources))
        version = ElasticBeanstalkAppVersionResource(
            context,
            app,
            self.elastic_beanstalk_bucket(),
            source_file.split("/")[-1],
            version)
        config = ElasticBeanstalkConfigTemplateResource(
            context,
            app,
            config_options,
            solution_stacks["docker"],
            "Pressurize docker elastic beanstalk config %s" % name,
            permission_resources=permission_resources
        )
        env = ElasticBeanstalkEnvResource(
            context,
            app,
            version,
            config,
            cname_prefix,
            "Pressurize env %s" % name
        )
        return [app, version, config, env] + bucket_resources

    def create_api_cluster(self, source_file, min_size=1, max_size=2):
        """
        Create a pressurize API cluster based on the Controller's config.

        InstanceType for this elasticbeanstalk deployment is determined by
        1) `instance_type` property
        2)  Defaults to t2.micro
        """

        cluster = Cluster(sanitize(self._cluster_name() + "api"), self._aws_context)
        version = str(random.randint(0, 100000))

        conf = self._controller.config
        config_options = {
            "aws:autoscaling:asg": {
                "MinSize": str(conf.get('api_min_size', min_size)),
                "MaxSize": str(conf.get('api_max_size', max_size))
            },
            "aws:autoscaling:launchconfiguration": {
                "InstanceType": conf.get('api_instance_type', 't2.micro')
            },
            "aws:elasticbeanstalk:command": {
                "DeploymentPolicy": "Immutable"
            },
            "aws:autoscaling:updatepolicy:rollingupdate": {
                "RollingUpdateType": "Immutable"
            },
            "aws:elasticbeanstalk:healthreporting:system": {
                "SystemType": "enhanced"
            },
            "aws:elasticbeanstalk:environment": {
                "EnvironmentType": "LoadBalanced"
            },
            "aws:elasticbeanstalk:cloudwatch:logs": {
                "StreamLogs": True,
                "DeleteOnTerminate": False,
                "RetentionInDays": 30
            }
        }
        resources = self.elastic_beanstalk_api_resources(version,
                                                         source_file,
                                                         config_options)
        for resource in resources:
            cluster.add_resource(resource)
        return cluster

    def read_instance_types(self, instance_types_file="instance_types.csv"):
        instance_types = []
        with open(instance_types_file, 'r') as f:
            x = csv.DictReader(f)
            for row in x:
                instance_types.append({
                    "name": row["Name"],
                    "memory": float(row["Memory"]),
                    "ecu": float(row["ECU"] if row["ECU"] != "Variable" else "0.5"),
                    "cost": float(row["Hourly Cost"][1:])
                })
        return instance_types

    def determine_instance_type(self, required_ecu, required_memory):
        """
        Find the least expensive EC2 instance that meets the provided
        ECU and memory requirements
        """
        if not hasattr(self, 'instance_types'):
            path = "/" + os.path.join(*(__file__.split("/")[:-1] + ["instance_types.csv"]))
            self.instance_types = self.read_instance_types(path)

        candidates = []
        for instance_idx in range(len(self.instance_types)):
            instance = self.instance_types[instance_idx]
            if instance['memory'] > required_memory and instance['ecu'] > required_ecu:
                candidates.append(instance)
        lwm = candidates[0]['cost']
        least_expensive = candidates[0]
        for candidate in candidates:
            if candidate['cost'] < lwm:
                least_expensive = candidate
                lwm = candidate['cost']
        return least_expensive

    def create_model_cluster(self, source_file, model_name, min_size=1, max_size=2):
        """
        Create a pressurize model cluster based on the Controller's config.

        InstanceType for this elasticbeanstalk deployment is determined by
         1) `required_ecu` and `required_memory` parameters.
         2) `instance_type` property
         3)  Defaults to t2.micro
        """

        cluster = Cluster(sanitize(self._cluster_name() + model_name), self._aws_context)
        version = str(random.randint(0, 100000))

        model_config = self._controller.models[model_name]

        # Determine instance type
        instance_type = 't2.micro'
        if 'required_ecu' in model_config:
            instance_type = self.determine_instance_type(
                model_config['required_ecu'],
                model_config['required_memory'])['name']
            print("Determined instance type: ", instance_type)
        if 'instance_type' in model_config:
            instance_type = model_config['instance_type']

        config_options = {
            "aws:autoscaling:asg": {
                "MinSize": str(model_config.get('min_size', min_size)),
                "MaxSize": str(model_config.get('max_size', max_size))
            },
            "aws:autoscaling:launchconfiguration": {
                "InstanceType": instance_type,
                "RootVolumeSize": str(model_config.get('storage_gb', 16))
            },
            "aws:elasticbeanstalk:command": {
                "DeploymentPolicy": "Immutable"
            },
            "aws:autoscaling:updatepolicy:rollingupdate": {
                "RollingUpdateType": "Immutable"
            },
            "aws:elasticbeanstalk:healthreporting:system": {
                "SystemType": "enhanced"
            },
            "aws:elasticbeanstalk:environment": {
                "EnvironmentType": "LoadBalanced"
            },
            "aws:elasticbeanstalk:cloudwatch:logs": {
                "StreamLogs": True,
                "DeleteOnTerminate": False,
                "RetentionInDays": 30
            }
        }
        resources = self.elastic_beanstalk_model_resources(model_name,
                                                           version,
                                                           source_file,
                                                           config_options)

        for resource in resources:
            cluster.add_resource(resource)
        return cluster

    def create_general_cluster(self):
        """
        Create a RedLeader cluster for AWS resource creation

        Not currently used.
        """
        cluster = Cluster(self._cluster_name(), self._aws_context)

        # Create a ref to cloudwatch logs so we can create an appropriate role
        logs = r.CloudWatchLogs(context)
        cluster.add_resource(logs)

        # Incorporate resources needed by individual models and our tables
        resources = self.create_bucket_resources() + \
                    self.default_dynamo_tables(context) + \
                    self.elastic_beanstalk_resources()
        for resource in resources:
            cluster.add_resource(resource)

        # Create our configuration bucket
        config_bucket = r.S3BucketResource(self.prefix_name("config"))
        cluster.add_resource(config_bucket)

        # Create a role for ELBS
        permissions = []
        for resource in resources:
            permissions.append(r.ReadWritePermission(resource))
        permissions.append(r.ReadWritePermission(logs))
        self.beanstalk_role = r.IAMRoleResource(context,
                                                permissions=permissions,
                                                services=["elasticbeanstalk.amazonaws.com"])
        cluster.add_resource(self.beanstalk_role)
        print(json.dumps(cluster.cloud_formation_template(), indent=4))
        return cluster



# Copyright 2016 Morgan McDermott & Blake Allen
"""

The Controller class coordinates the creation of API and Model
elastic beanstalk environments

"""
import os
import os.path
import json
import time
import datetime
import hashlib
import string
from zipfile import ZipFile
from threading import Thread
import importlib.util
import shutil
import random
import botocore
import subprocess

from redleader.managers import ElasticBeanstalkManager

import pressurize.ResourceManager as ResourceManager
import pressurize.AWSManager as AWSManager


class Controller(object):
    def __init__(self, config, aws_profile=None):
        self._defaults = {
        }
        self.validate_config(config)
        self.config = config
        self.local_queues = {}

        # Create map of models for efficient lookup
        self.models = {}
        for model in self.config['models']:
            self.models[model['name']] = model

        for key in self._defaults:
            if key not in config:
                setattr(self, key, self._defaults[key])

        self._aws_profile = aws_profile
        self._aws_manager = AWSManager.AWSManager(aws_profile=aws_profile)
        self._resource_manager = ResourceManager.ResourceManager(self)

    def create_resources(self, force_update=False):
        try:
            if force_update:
                self._cluster.blocking_delete(verbose=True)
            self._cluster.blocking_deploy(verbose=True)
        except Exception as e:
            if "AlreadyExists" not in "%s" % e:
                raise e
            else:
                print("Stack already exists %s" % e)

    def validate_config(self, config):
        required_keys = ['deployment_name', 'aws_region', 'models']
        accepted_keys = required_keys + ['api_min_size', 'api_max_size',
                                         'api_instance_type', 'custom_parameters']
        for key in required_keys:
            if key not in config:
                raise Exception('Config must have key %s' % key)
        for key in config:
            if key not in accepted_keys:
                raise Exception('Invalid key %s in top level config' % key)
        for model in config['models']:
            self.validate_model(model)

    def validate_model(self, model):
        required_keys = ['name', 'path', 'methods']
        for key in required_keys:
            if key not in model:
                raise Exception('Model %s config must have key %s' %
                                (model.get('name', 'unnamed'), key))
        accepted_keys = required_keys + ['min_size', 'max_size', 'instance_type',
                                         'required_memory', 'required_ecu',
                                         'required_resources', 'storage_gb',
                                         'custom_parameters'
        ]
        for key in model:
            if key not in accepted_keys:
                raise Exception('Invalid key %s in model %s config' %
                                (key, model['name']))


    def recursively_add_files_to_zip(self, source_path, zipfile, base=""):
        for filename in os.listdir(source_path):
            f = os.path.join(source_path, filename)
            if os.path.islink(f):
                continue
            if filename[0] == "." or ".zip" in filename:
                continue
            elif os.path.isdir(f):
                self.recursively_add_files_to_zip(f, zipfile,
                                                  os.path.join(base, filename))
            else:
                zipfile.write(f, os.path.join(base, filename))

    def custom_config(self, model_name):
        """
        Returns a copy of the current config, excluding all models
        except the given one. Used for deployment to individual model servers.
        """
        new_config = {}
        for k in self.config:
            new_config[k] = self.config[k]
        new_config['models'] = list(filter(lambda x: x['name'] == model_name, new_config['models']))
        return new_config

    def create_api_package(self):
        """
        Creates a zip package deployable to elastic beanstalk for the pressurize API
        """
        template_dir = "/" + os.path.join(*(__file__.split("/")[:-1] + ["api"]))
        filename = "deploy_api.zip"
        target = os.path.join(os.getcwd(), filename)
        try:
            os.remove(target)
        except FileNotFoundError:
            pass

        with ZipFile(target, 'w') as zipfile:
            self.recursively_add_files_to_zip(template_dir, zipfile)

            # Write the custom config for this particular model server
            tmpfile = '/tmp/custom_api_config.json'
            with open(tmpfile, 'w') as f:
                json.dump(self.config, f)
            zipfile.write(tmpfile, 'pressurize.json')
        return filename

    def create_model_package(self, source_path, model_name):
        """
        Creates a zip package deployable to elastic beanstalk for the given model
        """
        template_dir = "/" + os.path.join(*(__file__.split("/")[:-1] + ["model", "deploy_template"]))
        filename = "deploy_model_%s.zip" % model_name
        target = os.path.join(os.getcwd(), filename)
        try:
            os.remove(target)
        except FileNotFoundError:
            pass

        with ZipFile(target, 'w') as zipfile:
            self.recursively_add_files_to_zip(template_dir, zipfile)
            self.recursively_add_files_to_zip(source_path, zipfile)

            # Write the custom config for this particular model server
            custom_config = self.custom_config(model_name)
            tmpfile = '/tmp/custom_config.json'
            with open(tmpfile, 'w') as f:
                json.dump(custom_config, f)
            zipfile.write(tmpfile, 'pressurize.json')
        return filename

    def run_local(self, model_name, source_path=None, port='5000', docker=True):
        source_path = source_path or os.getcwd()

        local_path = os.path.join("/tmp/", "pressurize_local_"+model_name+str(random.randint(0, 10000)))
        print("Local Run: Creating temp deployment ", local_path)

        template_dir = "/" + os.path.join(*(__file__.split("/")[:-1] + ["model", "deploy_template"]))
        copytree(template_dir, local_path)
        copytree(source_path, local_path)
        os.chmod(os.path.join(local_path, "pressurize.json"), 777)
        os.remove(os.path.join(local_path, "pressurize.json"))

        custom_config = self.custom_config(model_name)
        with open(os.path.join(local_path, "pressurize.json"), 'w') as f:
            json.dump(custom_config, f)

        # Load as module if docker is false
        if docker == False:
            print("Loading model as module")
            spec = importlib.util.spec_from_file_location("server", os.path.join(local_path, "server.py"))
            server = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(server)
            server.main(port=port)
            return

        # Copy AWS Credentials
        print("Loading model in docker container")
        print("Expanded user", os.path.expanduser("~"))
        shutil.copy2(os.path.join(os.path.expanduser("~"), ".aws/credentials"),
                     os.path.join(local_path, "credentials"))

        # Run the docker build script
        cmd = " ".join(["/".join(__file__.split("/")[:-1] + ["model", "run_local.sh"]),
                                       local_path,
                                       model_name.lower(),
                                       str(port)])
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                  stderr=subprocess.STDOUT, shell=True)
        for line in p.stdout:
            print(line.decode('utf-8').replace("\n", ""))

    def deploy_api(self, dry_run=False):
        """
        Deploys an elastic beanstalk environment for the pressurize API
        """
        packagefile = self.create_api_package()
        print("Created API elastic beanstalk package ", packagefile)
        bucket_name = self._resource_manager.elastic_beanstalk_bucket()
        manager = ElasticBeanstalkManager(self._aws_manager)
        manager.upload_package(bucket_name, os.path.join(os.getcwd(), packagefile), packagefile)
        cluster = self._resource_manager.create_api_cluster(packagefile)
        if dry_run:
            return cluster.cloud_formation_template()
        try:
            cluster.blocking_deploy(verbose=True)
        except botocore.exceptions.ClientError as e:
            if "exists" in "%s" % e:
                print("Cluster already exists. Updating")
                cluster.blocking_update(verbose=True)
            else:
                raise e

    def deploy_model(self, source_path, model_name, blocking=True, dry_run=False):
        """
        Deploys an elastic beanstalk environment for the given model
        """
        packagefile = self.create_model_package(source_path, model_name)
        print("Created model elastic beanstalk package ", source_path, packagefile)
        bucket_name = self._resource_manager.elastic_beanstalk_bucket()
        manager = ElasticBeanstalkManager(self._aws_manager)
        manager.upload_package(bucket_name, os.path.join(os.getcwd(), packagefile), packagefile)
        cluster = self._resource_manager.create_model_cluster(packagefile, model_name)
        if dry_run:
            return cluster.cloud_formation_template()
        try:
            if blocking:
                cluster.blocking_deploy()
            else:
                cluster.deploy()
        except botocore.exceptions.ClientError as e:
            if "exists" in "%s" % e:
                print("Cluster already exists. Updating")
                if blocking:
                    cluster.blocking_update(verbose=True)
                else:
                    cluster.update()
            else:
                raise e

    def deploy_models(self, source_path=None, dry_run=False):
        if source_path is None:
            source_path = os.getcwd()
        res = []
        for model in self.models:
            print("Deploying model %s" % model)
            res.append(self.deploy_model(source_path, model, blocking=False, dry_run=dry_run))
        return res

    def destroy_api_cluster(self):
        """
        Destroys the elastic beanstalk environment for the given model
        """
        print("Destroying API cluster...")
        filename = "deploy_api.zip"
        cluster = self._resource_manager.create_api_cluster(filename)
        cluster.blocking_delete(verbose=True)
        print("Destroyed API cluster")

    def create_token(self, token_lifetime_in_hours):
        chars = (string.ascii_letters + string.digits + '!@#$%^&*()')[:64]
        token_key = "".join([chars[os.urandom(1)[0] % len(chars)] for x in range(32)])
        secret = "".join([chars[os.urandom(1)[0] % len(chars)] for x in range(32)])

        ddb = self._aws_manager.get_client('dynamodb')
        expires = int(time.time() + token_lifetime_in_hours * 60 * 60)
        res = ddb.put_item(
            TableName=self._resource_manager.prefix_name("auth"),
            Item={
                "token_key": {"S": token_key},
                "secret": {"S": secret},
                "expires": {"N": str(expires)}
            }
        )
        return {
            "token_key": token_key,
            "secret": secret,
            "expires": expires
        }

    def destroy_model_cluster(self, model_name):
        """
        Destroys the elastic beanstalk environment for the given model
        """
        filename = "deploy_model_%s.zip" % model_name
        cluster = self._resource_manager.create_model_cluster(filename, model_name)
        cluster.blocking_delete(verbose=True)
        print("Destroyed cluster for model " % model_name)

class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return int(time.mktime(obj.timetuple()))
        return json.JSONEncoder.default(self, obj)

def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            try:
                os.makedirs("/" + "/".join(d.split("/")[:-1]))
            except FileExistsError:
                pass
            if os.path.exists(d):
                os.chmod(d, 644)
            shutil.copy2(s, d)



"""
 model_server.py
 Runs a simple API granting access to all available models.
 Usage: python model_server.py

 This server uses a separate process to perform predictions, with a pipe
 from `multiprocessing` connecting the flask server thread to the prediction
 process. Access to the server side of the pipe is guarded by a Lock from multiprocessing.
"""


from multiprocessing import Queue, Process, Pipe, Lock, log_to_stderr
import logging
import logging.handlers

import multiprocessing
import queue as Q
import random
import os.path
import json
import os
import importlib

from flask import Flask, jsonify, request, abort, make_response
from flask_cors import CORS, cross_origin
import tensorflow as tf
import boto3
import botocore


class ModelServer(object):
    def __init__(self, config, source_path, resource_path, model_conf, pipe):
        self._pipe = pipe
        self._resources = ModelServer.acquire_resources(config, model_conf, resource_path)
        self._model_class = ModelServer.import_model(model_conf['path'], source_path)
        self._logger = self.setup_logging()
        self._model = self._model_class(self._resources, config=model_conf, logger=self._logger)

    def setup_logging(self):
        handler = logging.handlers.WatchedFileHandler(
            os.environ.get("PRESSURIZE_LOGFILE", "/var/log/pressurize.log"))
        formatter = logging.Formatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        logger = logging.getLogger()
        logger.setLevel(os.environ.get("LOGLEVEL", "INFO"))
        logger.addHandler(handler)
        return logger

    def run(self):
        #logger = multiprocessing.log_to_stderr()
        self._logger.info('About to enter model processing loop')
        try:
            with self._model.modelcontext():
                while True:
                    item = self._pipe.recv()
                    if not hasattr(self._model, item['method']):
                        self._pipe.send({"error": "Model does not have method %s" % item['method']})
                    try:
                        preprocessed = self._model.preprocess(item["data"])
                        result = getattr(self._model, item['method'])(preprocessed)
                        self._pipe.send({"result": result})
                    except Exception as e:
                        self._pipe.send({"error": "Exception: " + str(e)})
                        self._logger.exception("Encountered error during invocation of method %s: %s" %
                                     (item['method'], str(e)))
        except Exception as e:
            self._logger.exception("Unexpected error encountered during model setup or request processing.")

    @staticmethod
    def import_model(path, source_path):
        """
        Imports a PressurizeModel given a pressurize.json model config path
        e.g.) given the path "TestModel.TestModel", imports:
              import TestModel from TestModel.TestModel
        """
        filepath = os.path.join(*path.split('.')) + '.py'
        fullpath = os.path.join(source_path, filepath)
        spec = importlib.util.spec_from_file_location(path, fullpath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return getattr(module, path.split(".")[-1])

    @staticmethod
    def acquire_resources(config, model, model_resource_path):
        session = boto3.Session(region_name=config['aws_region'])
        client = session.client('s3')

        resources = {}
        for resource_name in model['required_resources']:
            s3_path = model['required_resources'][resource_name]
            print("Downloading resource %s from %s" % (resource_name, s3_path))
            parts = s3_path.split("/")
            if len(parts) < 4:
                raise RuntimeError("Invalid s3 resource in config: " + resource_name)
            if parts[0] != "s3:":
                resources[resource_name] = s3_path
                continue

            bucket = parts[2] # s3://{bucket}
            key = "/".join(parts[3:]) # s3://{bucket}/{key/with/slashes}

            # Ensure local folder exists
            local_folder = model_resource_path
            if len(parts) != 4:
                local_folder = os.path.join(local_folder, *parts[3:-1])
            if not os.path.exists(local_folder):
                os.makedirs(local_folder)

            local_path = os.path.join(model_resource_path, *parts[3:])
            try:
                client.download_file(bucket, key, local_path)
            except botocore.exceptions.ClientError as e:
                raise RuntimeError("Failed to download resource '%s' @ %s: %s" % \
                      (resource_name, s3_path, str(e)))
            resources[resource_name] = local_path
        return resources

serverPipeLock = Lock()
pipes = {}
app = Flask(__name__)
CORS(app)

@app.route('/api/<string:model>/<string:method>/', methods=['POST'])
def executeModelMethod(model, method):
    data = request.get_json()
    if data is None:
        print("Data not provided")
        return make_response(jsonify({'error': 'Data not provided'}), 400)

    if model not in pipes:
        print("Error: Model does not exist")
        return make_response(jsonify({'error': 'Model does not exist'}), 404)

    # Send our task over the pipe and wait for a result.
    resId = str(random.random())
    serverPipeLock.acquire()
    serverPipe = pipes[model][0]
    serverPipe.send({
        "model": model,
        "method": method,
        "data": data,
        "resId": resId
    })
    res = serverPipe.recv()
    serverPipeLock.release()

    if "error" in res:
        return make_response(jsonify({'error': res['error']}), 400)
    return jsonify(res)

def run_server(config, source_path=os.getcwd(), resource_path=os.getcwd(),
               port='5000', debug=False, separate_process=False):
    """
    run_server takes a list of models from a pressurize.json config,
    starting a ModelServer in a separate process for each model.
    """
    for model in config['models']:
        model_resource_path = os.path.join(resource_path, "resources", model['name'])
        pipes[model['name']] = Pipe()
        model_server = ModelServer(config, source_path, model_resource_path,
                                   model, pipes[model['name']][1])

        taskProcess = Process(target=model_server.run)
        taskProcess.start()
    if separate_process:
        serverProcess = Process(target=app.run, args=(('0.0.0.0', port, debug)))
        serverProcess.start()
        return serverProcess
    else:
        app.run(host='0.0.0.0', port=port, debug=debug)



# Copyright 2016 Morgan McDermott & Blake Allen
import botocore
import boto3

class AWSManager(object):
    def __init__(self, aws_region="us-west-2", aws_profile=None,
                 aws_access_key_id=None, aws_secret_access_key=None):
        self._aws_region = aws_region
        self._aws_profile = aws_profile
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._session = None
        self.create_session()
        self.clients = {}

    def create_session(self):
        if self._session is None:
            try:
                print("Creating AWS Session with profile %s" % self._aws_profile)
                self._session = boto3.Session(profile_name=self._aws_profile,
                                              region_name=self._aws_region
                )
            except botocore.exceptions.ProfileNotFound:
                print("Profile not found. Attempting to utilize provided keys")
                self._session = boto3.Session(
                    region_name=self._aws_region,
                    aws_access_key_id=self._aws_access_key_id,
                    aws_secret_access_key=self._aws_secret_access_key
                )
            except Exception as e: # TODO: Determine specific exception
                self._session = boto3.Session(region_name=self._aws_region)

    def get_client(self, service):
        """
        Return a client configured with current credentials and region
        """
        if service not in self.clients:
            self.clients[service] = self._session.client(service)
        return self.clients[service]



from contextlib import contextmanager
import tensorflow as tf

class PressurizeModel(object):
    def __init__(self, resources, config=None, logger=None):
        self.resources = resources
        self.config = config

    @contextmanager
    def modelcontext(self):
        """
        Default context manager does nothing
        """
        yield

    def preprocess(self, data):
        return data

    def predict(self, data):
        raise NotImplementedError

class PressurizeTFModel(PressurizeModel):
    def tf_setup(self):
        raise NotImplementedError

    @contextmanager
    def modelcontext(self, device="/cpu:0"):
        """
        Tensorflow context manager creates a new graph and executes code
        in the context of a default graph and the given tensorflow device
        """
        graph = tf.Graph()
        with graph.as_default():
            with tf.device(device):
                self.tf_setup()
                yield



#!/usr/bin/env python
from setuptools import setup, find_packages

install_requires = [
    'click',
    'boto3',
    'requests'
]

setup(
    name='pressurize',
    version='0.15',
    description='Cloud Machine Learning Deployment',
    author='Morgan McDermott',
    author_email='morganmcdermott@gmail.com',
    long_description='',
    url='https://github.com/mmcdermo/pressurize',
    license='All rights reserved',
    zip_safe=False,
    keywords='',
    packages=find_packages(),
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'pressurize = pressurize.cli:main',
        ]
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ]
)



import os
import os.path
import json
import pressurize.model.model_server as server

def main(port='5000'):
    model_dir = "/" + os.path.join(*__file__.split("/")[:-1])
    config = {}
    with open(os.path.join(model_dir, "pressurize.json")) as f:
        config = json.load(f)
    print("Running model server on port %s" % port)
    process = server.run_server(config,
                                source_path=os.path.join(model_dir),
                                port=port)

if __name__ == '__main__':
    main()



import os
import json
import os.path
import click
import pressurize
from pressurize import Controller

import botocore

@click.group()
@click.version_option(version=pressurize.__version__, message='%(prog)s %(version)s')
@click.option('--debug/--no-debug', default=False,
              help='Write debug logs to standard error.')
@click.pass_context
def cli(ctx, debug=False):
    if not ctx.obj:
        ctx.obj = {}
    ctx.obj['config_file'] = 'pressurize.json'
    ctx.obj['project_dir'] = os.getcwd()
    ctx.obj['debug'] = debug

@cli.command()
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.pass_context
def deploy(ctx, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
       click.echo('Error with config: %s' % e)
       raise click.Abort()

    # Deploy API
    try:
        controller.deploy_api()
    except botocore.exceptions.ClientError as e:
        print("Failed to deploy API. %s" % e)

    # Deploy Models
    try:
        controller.deploy_models()
    except botocore.exceptions.ClientError as e:
        print("Failed to deploy models. %s" % e)


@cli.command(name="deploy-model")
@click.argument("model_name")
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.pass_context
def deploy_model(ctx, model_name, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
       click.echo("Error with config: %s" % e)
       raise click.Abort()

    if model_name not in controller.models:
        click.echo("Model %s not found in config")
        raise click.Abort()

    # Deploy Model
    source_path = os.getcwd()
    try:
        controller.deploy_model(source_path, model_name)
    except botocore.exceptions.ClientError as e:
        print("Failed to deploy model %s. %s" % (model_name, e))

@cli.command(name="deploy-models")
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.pass_context
def deploy_models(ctx, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
       click.echo("Error with config: %s" % e)
       raise click.Abort()

    # Deploy Models
    try:
        controller.deploy_models()
        print("------------------------")
        print("Models are being updated.")
    except botocore.exceptions.ClientError as e:
        print("Failed to deploy models. %s" % e)

@cli.command(name="deploy-api")
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.pass_context
def deploy_api(ctx, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
       click.echo("Error with config: %s" % e)
       raise click.Abort()

    # Deploy Models
    try:
        controller.deploy_api()
    except botocore.exceptions.ClientError as e:
        print("Failed to deploy API. %s" % e)

@cli.command()
@click.pass_context
@click.argument("model_name")
@click.option("--port", default=5001)
@click.option("--docker/--no-docker", default=True)
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
def local(ctx, model_name, port, docker, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
        click.echo('Error with config: %s' % e)
        raise click.Abort()
    controller.run_local(model_name, port=port, docker=docker)

@cli.command(name="dry-run")
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.pass_context
def dry_run(ctx, aws_profile):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
        click.echo('Error with config: %s' % e)
        raise click.Abort()
    print(json.dumps(controller.deploy_api(dry_run=True), indent=4))
    print(json.dumps(controller.deploy_models(dry_run=True), indent=4))

@cli.command(name="create-token")
@click.option('--aws-profile', default=None,
              help='AWS Profile to use for cluster commands')
@click.option("--token-lifetime-in-hours", default=24*30*12)
@click.pass_context
def create_token(ctx, aws_profile, token_lifetime_in_hours):
    if ctx.obj['config_file'] not in os.listdir(ctx.obj['project_dir']):
        click.echo('No pressurize.json file found in directory')
        raise click.Abort()

    print("Config path", os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']))
    with open(os.path.join(ctx.obj['project_dir'], ctx.obj['config_file']), 'r') as f:
        config = json.load(f)
    try:
        controller = Controller.Controller(config, aws_profile=aws_profile)
    except Exception as e:
        click.echo('Error with config: %s' % e)
        raise click.Abort()

    token = controller.create_token(token_lifetime_in_hours)
    print("----- Generated Token ------")
    print(json.dumps(token, indent=4))


def main():
    cli(obj={})
